/* Copyright 2004 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: GameState.cpp,v 1.9 2004/10/17 23:34:05 tonic Exp $
 * $Revision: 1.9 $
 */

#include <math.h>
#include "main.h"


static const UINT32 sWanderOffsets[8] = {
    GameState::MAP_SIZE - GameState::MAP_WIDTH - 1,
    GameState::MAP_SIZE - GameState::MAP_WIDTH,
    GameState::MAP_SIZE - GameState::MAP_WIDTH + 1,
    GameState::MAP_SIZE - 1,
    1,
    GameState::MAP_WIDTH - 1,
    GameState::MAP_WIDTH,
    GameState::MAP_WIDTH + 1
};


GameState::GameState(UIGraphics *graphics) :
    mInitialized(FALSE),
    //mGraphics(graphics),
    mGameState(STATE_INIT),
    mGLInitialized(FALSE)
{
    mFPS = 60;

    //assert(graphics != NULL);
    //mUI.setGraphics(mGraphics);
}


GameState::~GameState()
{
    deinit();
    deinitUI();
    if (mInitialized)
    {
        //mUI.deleteAllUIComponents();
        freeImage(mQuadCircleImage);
        freeImage(mBladeImage);
        freeImage(mBulletImage);
    }
}


BOOL GameState::init()
{
    mInitialized = TRUE;

    mQuadCircleImage = loadImage(DATA_PREFIX "quadcircle.png");
    mBladeImage = loadImage(DATA_PREFIX "blade.png");
    mBulletImage = loadImage(DATA_PREFIX "bullet.png");

    if (mInitialized)
        mInitialized = initUI();

    return mInitialized;
}


void GameState::deinit()
{
    UINT a;
    for (a = 0; a < mDLATextureCount; ++a)
        glDeleteTextures(1, &mDLATextures[a]);
}


BOOL GameState::initUI()
{
    GameApp *ga = GameApp::getInstance();
    INT scrW = ga->getScreenWidth(), scrH = ga->getScreenHeight();

    mSmallFont = ga->getSmallFont();
    mMediumFont = ga->getMediumFont();
    mBigFont = ga->getBigFont();
    mNumFont = ga->getNumFont();

    return TRUE;
}


void GameState::deinitUI()
{
    //mUI.deleteAllUIComponents();
}


void GameState::onActivate(UINT32 time)
{
    if (!mInitialized)
    {
        BOOL res = init();
        assert(res);
    }

    if (!mGLInitialized)
    {
        mDLATextureCount = 0;
        UINT w = MAP_IMAGE_WIDTH, h = MAP_IMAGE_HEIGHT;
        void *texmem = calloc(w * h * 3, 1);
        if (texmem == NULL)
            mGLInitialized = FALSE;
        else
        {
            while (w > 4 && h > 4)
            {
                glGenTextures(1, &mDLATextures[mDLATextureCount]);
                glBindTexture(GL_TEXTURE_2D, mDLATextures[mDLATextureCount]);
                //glTexImage2D(GL_TEXTURE_2D, 0, 3, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, texmem);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                w /= 2;
                h /= 2;
                ++mDLATextureCount;
            }
            free(texmem);
        }
        glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
        glEnable(GL_LINE_SMOOTH);
        mGLInitialized = TRUE;
    }

    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);

    mDLAPCount = DLAP_COUNT;
    mScore = 0;
    mLevel = 1;

    startGame();
}


void GameState::onDeactivate()
{
    deinitGame();
}


void GameState::startGame()
{
    UINT32 a;

    mFrame = 0;

    mRandom.init_genrand(SDL_GetTicks());

    mMap = new UINT32[MAP_SIZE];
    for (a = 0; a < MAP_SIZE; ++a)
        mMap[a] = 0;
    mMap[(MAP_HEIGHT / 2) * MAP_WIDTH + (MAP_WIDTH / 2)] = 1;
    mStuckDLAPs = 1;
    /*
    mMapImage = SDL_CreateRGBSurface(SDL_SWSURFACE, MAP_IMAGE_WIDTH, MAP_IMAGE_HEIGHT,
                                     32, 0xff, 0xff00, 0xff0000, 0xff000000);
    assert(mMapImage != NULL);
    SDL_FillRect(mMapImage, NULL, 0xff000000);
    */

    mDLAPInitRadius = DLA_MIN_INIT_RADIUS;
    mDLAPs = new DLAParticle[mDLAPCount];
    for (a = 0; a < mDLAPCount; ++a)
        initDLAParticle(a);

    mBorderR = (FLOAT)(mRandom.genrand_real1() * 0.15 + 0.15);
    mBorderG = (FLOAT)(mRandom.genrand_real1() * 0.15 + 0.15);
    mBorderB = (FLOAT)(mRandom.genrand_real1() * 0.2 + 0.2);

    // clear screen
    SDL_FillRect(SDL_GetVideoSurface(), NULL, 0xff000000);
    /*
    SDL_Rect vlineRect = { 512, 0, 1, 480 };
    SDL_FillRect(SDL_GetVideoSurface(), &vlineRect, 0xff806040);
    */

    makeDLAPalette();

    mBullets.clear();
    mBranchErasers.clear();

    mBladeAngle = mBladeDestAngle = PI / 4;
    mKeyLeft = mKeyRight = FALSE;
    mBladeSpeed = 0.015f;
    mFireDelay = 0;
    mFireDelayAdder = FIRE_DELAY;
    mBulletCount = 5 + mLevel * 5;
    mEraserComboLength = 0;
    mEraserComboLengthDisplay = mEraserComboLengthDisplayTicks = 0;
    mBulletRegenCount = BULLET_REGEN_RATE;

    setGameState(STATE_RUNNING);

    playSoundEffect(SND_LEVEL_BEGIN);
}


void GameState::deinitGame()
{
    delete[] mDLAPs;
    delete[] mMap;
}


void GameState::restartGame()
{
    deinitGame();

    mDLAPCount += DLAP_COUNT_LEVEL_ADD;
    if (mDLAPCount > DLAP_COUNT_MAX)
        mDLAPCount = DLAP_COUNT_MAX;

    ++mLevel;

    startGame();
}


void GameState::onKeyDown(const SDLKey &key)
{
    if (key == SDLK_SPACE)
    {
        if (mFireDelay == 0 && getGameState() == STATE_RUNNING)
            fire();
        else if (getGameState() == STATE_GAME_OVER && mFrame > 30)
            GameApp::getInstance()->setState(GameApp::STATE_TITLE);
        else if (getGameState() == STATE_LEVEL_CHANGE && mFrame > 30)
            restartGame();
    }
    else if (key == SDLK_LEFT || key == SDLK_KP4)
        mKeyLeft = TRUE;
    else if (key == SDLK_RIGHT || key == SDLK_KP6)
        mKeyRight = TRUE;
}


void GameState::onKeyUp(const SDLKey &key)
{
    if (key == SDLK_LEFT || key == SDLK_KP4)
        mKeyLeft = FALSE;
    else if (key == SDLK_RIGHT || key == SDLK_KP6)
        mKeyRight = FALSE;
}


BOOL GameState::update(UINT32 time)
{
    mTime = time;

    ++mFrame;

    switch (getGameState())
    {
    default:
        break;

    case STATE_RUNNING:
        updateGameRunningState();
        break;

    case STATE_LEVEL_CHANGE:
        if (mFrame == 1)
        {
            mDisplayLevelBonusScore = mBulletCount * 10;
            mScore += mDisplayLevelBonusScore;
        }
        break;

    case STATE_GAME_OVER:
        break;
    }

    if (mEraserComboLengthDisplayTicks > 0)
        --mEraserComboLengthDisplayTicks;

    return TRUE;
}


void GameState::updateGameRunningState()
{
    int a;
    for (a = 0; a < DLA_ITERATIONS; ++a)
        diffusionLimitedAggregation();

    if (mKeyLeft)
        mBladeDestAngle += mBladeSpeed;
    if (mKeyRight)
        mBladeDestAngle -= mBladeSpeed;

    for (a = 0; a < BULLET_UPDATE_ITERATIONS; ++a)
        updateBullets();
    updateBranchErasers();

    mBladeAngle = 0.85f * mBladeAngle + 0.15f * mBladeDestAngle;

    if (mFireDelay > 0)
        --mFireDelay;

    if (--mBulletRegenCount == 0)
    {
        mBulletRegenCount = BULLET_REGEN_RATE;
        ++mBulletCount;
    }

    //if ((mFrame & 3) == 0)
    --mDLAPInitRadius;
    if (mDLAPInitRadius < DLA_MIN_INIT_RADIUS)
        mDLAPInitRadius = DLA_MIN_INIT_RADIUS;

    if (((mFrame & 7) == 0) &&
        mBranchErasers.size() > 0)
    {
        playSoundEffect(SND_ERASER);
    }
}


void GameState::render(SDL_Surface *screen)
{
    clear(0, 0, 0);

    drawDLA();
    drawHUD();

    //mUI.render();
}


void GameState::makeDLAPalette()
{
    INT32 comps[3] = { 0, 0, 0 };
    UINT a, comp, comp2, stride, col;
    const UINT stridesGen = DLA_PALETTE_STRIDES_GEN_MIN +
        mRandom.genrand_int32() % (DLA_PALETTE_STRIDES_GEN_MAX - DLA_PALETTE_STRIDES_GEN_MIN + 1);
    for (a = 0; a < stridesGen; ++a)
    {
        do comp = mRandom.genrand_int32() % 3; while (comps[comp] == 255);
        comps[comp] += DLA_PALETTE_STRIDE_LENGTH;
        if (comps[comp] > 255)
            comps[comp] = 255;
    }
    col = 0;
    for (stride = 0; stride < DLA_PALETTE_STRIDES; ++stride)
    {
        do comp = mRandom.genrand_int32() % 3; while (comps[comp] == 0);
        comp2 = mRandom.genrand_int32() % 3;
        for (a = 0; a < DLA_PALETTE_STRIDE_LENGTH; ++a)
        {
            if (--comps[comp] < 0)
                comps[comp] = 0;
            if (stride >= DLA_PALETTE_STRIDES / 2 && comp != comp2 && comps[comp2] < 255)
                ++comps[comp2];
            assert(comps[0] >= 0 && comps[0] <= 255);
            assert(comps[1] >= 0 && comps[1] <= 255);
            assert(comps[2] >= 0 && comps[2] <= 255);
            mDLAPalette[col++] = 0xff000000 | (comps[0] << 16) | (comps[1] << 8) | (comps[2]);
        }
    }
}


void GameState::drawDLA()
{
    UINT32 x, y, imageOffset, mapOffset = 0;

    //SDL_Surface *mapImage = mMapImage;
    SDL_Surface *mapImage = SDL_GetVideoSurface();
    INT lockResult = SDL_LockSurface(mapImage);
    assert(lockResult == 0);

    // draw image to lower left corner for OpenGL texture copy
    imageOffset = (mapImage->h - MAP_HEIGHT) * mapImage->pitch;

    // note that by offset y goes actually [4..h-4].
    for (y = 0; y < MAP_HEIGHT; ++y)
    {
        for (x = 0; x < MAP_WIDTH; ++x)
        {
            //((UINT32 *)mMapImage->pixels)[imageOffset] = 0xff000000 | (mMap[mapOffset] * 0xffffff);
            if (mMap[mapOffset])
            {
                UINT32 mapValue = mMap[mapOffset];
                if (mapValue >= DLA_PALETTE_LENGTH)
                    mapValue = DLA_PALETTE_LENGTH - 1;
                UINT32 color = mDLAPalette[mapValue];
                //color = ((0xff - color) << 8) | color | 0xff000000;
                const UINT8 *destPix = (UINT8 *)mapImage->pixels + imageOffset;
                *((UINT32 *)destPix) = color;
            }
            else
            {
                const UINT8 *destPix = (UINT8 *)mapImage->pixels + imageOffset;
                *((UINT32 *)destPix) = 0;
            }

            imageOffset += 4;
            ++mapOffset;
        }
        //imageOffset += 1 * MAP_IMAGE_WIDTH;
        imageOffset += mapImage->pitch - MAP_WIDTH * 4;
    }
    SDL_UnlockSurface(mapImage);

    /*
    SDL_Rect srcRect;
    srcRect.x = 0;
    srcRect.y = 16;
    srcRect.w = 512;
    srcRect.h = 480;
    */
    //INT blitRes = SDL_BlitSurface(mMapImage, NULL, SDL_GetVideoSurface(), NULL);
    //assert(blitRes == 0);
    //SDL_UpdateRect(SDL_GetVideoSurface(), 0, 0, 512, 480);
    // image is updated to lower left corner
    SDL_UpdateRect(SDL_GetVideoSurface(), 0, mapImage->h - MAP_HEIGHT, MAP_WIDTH, MAP_HEIGHT);

    enhanceDrawnDLA();
}


void GameState::enhanceDrawnDLA()
{
    enter2DMode(WIDTH, HEIGHT);
    glDisable(GL_LIGHTING);
    glDisable(GL_BLEND);

    glBindTexture(GL_TEXTURE_2D, mDLATextures[0]);
    //glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, MAP_WIDTH, MAP_HEIGHT);
    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, MAP_WIDTH, MAP_HEIGHT, 0);
    glColor4f(1, 1, 1, 1);

    int w = MAP_WIDTH / 2, h = MAP_HEIGHT / 2, a;
    for (a = 1; a < mDLATextureCount; ++a)
    {
        glBindTexture(GL_TEXTURE_2D, mDLATextures[a - 1]);
        glBegin(GL_QUADS);
        const UINT yAdd = HEIGHT - h;
        glTexCoord2f(0, 1); glVertex2f(0, yAdd);
        glTexCoord2f(1, 1); glVertex2f(w, yAdd);
        glTexCoord2f(1, 0); glVertex2f(w, yAdd + h);
        glTexCoord2f(0, 0); glVertex2f(0, yAdd + h);
        glEnd();
        glBindTexture(GL_TEXTURE_2D, mDLATextures[a]);
        glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, w, h, 0);
        w /= 2;
        h /= 2;
    }

    glClearColor(0, 0, 0, 1);
    glClear(GL_COLOR_BUFFER_BIT);
    glViewport(0, 0, WIDTH, HEIGHT);

    // glow
    UINT tex;
    const FLOAT a1 = (FLOAT)atan2(HEIGHT / 2, MAP_WIDTH);
    const FLOAT a2 = (FLOAT)atan2(HEIGHT / 2, -MAP_WIDTH);
    const FLOAT a3 = (FLOAT)atan2(-HEIGHT / 2, -MAP_WIDTH);
    const FLOAT a4 = (FLOAT)atan2(-HEIGHT / 2, MAP_WIDTH);
    for (tex = 0; tex < mDLATextureCount; ++tex)
    {
        FLOAT g = 1;

        if (tex > 0)
            g = 0.7f + 0.3f - (0.3f * tex / mDLATextureCount);
        glBindTexture(GL_TEXTURE_2D, mDLATextures[tex]);
        glColor4f(g, g, g, 1);
        glBegin(GL_TRIANGLE_STRIP);
        const FLOAT vGap = (FLOAT)(MAP_HEIGHT * 2 - HEIGHT) / 2 / (MAP_HEIGHT * 2);
        /*
        glTexCoord2f(0, 1 - vGap); glVertex2f(0, 0);
        glTexCoord2f(1, 1 - vGap); glVertex2f(MAP_WIDTH * 2, 0);
        glTexCoord2f(0, vGap); glVertex2f(0, HEIGHT);
        glTexCoord2f(1, vGap); glVertex2f(MAP_WIDTH * 2, HEIGHT);
        */
        FLOAT t = (FLOAT)(sin(mTime * 0.002f) * 0.005);
        const FLOAT d = (FLOAT)sqrt((MAP_WIDTH + 2) * (MAP_WIDTH + 2) + ((HEIGHT / 2) + 2) * ((HEIGHT / 2) + 2));
        const FLOAT xAdd = MAP_WIDTH + DLA_X_OFFSET - 1;
        glTexCoord2f(0, 1 - vGap);
        glVertex2f(cos(t + a2) * d + xAdd, sin(t + a2) * d + HEIGHT / 2);
        glTexCoord2f(1, 1 - vGap);
        glVertex2f(cos(t + a1) * d + xAdd, sin(t + a1) * d + HEIGHT / 2);
        glTexCoord2f(0, vGap);
        glVertex2f(cos(t + a3) * d + xAdd, sin(t + a3) * d + HEIGHT / 2);
        glTexCoord2f(1, vGap);
        glVertex2f(cos(t + a4) * d + xAdd, sin(t + a4) * d + HEIGHT / 2);
        glEnd();

        if (tex == 0)
        {
            glEnable(GL_BLEND);
            glBlendFunc(GL_ONE, GL_ONE);
        }
    }

    glDisable(GL_BLEND);
    leave2DMode();
}


void GameState::addBranchEraser(UINT32 offset)
{
    if (mBranchErasers.find(offset) == mBranchErasers.end() &&
        mNewBranchErasers.find(offset) == mNewBranchErasers.end())
        mNewBranchErasers.insert(offset);
}


void GameState::updateBranchErasers()
{
    UINT oldEraserCount = mBranchErasers.size();

    BranchEraserSet::iterator it = mBranchErasers.begin();
    BranchEraserSet::iterator itEnd = mBranchErasers.end();
    while (it != itEnd)
    {
        UINT32 offset = *it;
        INT32 branchValue = (INT32)mMap[offset];
        if (mMap[offset])
            --mStuckDLAPs;
        mMap[offset] = 0;
        ++mScore;
        UINT a;
        for (a = 0; a < 8; ++a)
        {
            UINT32 testOffset = (offset + sWanderOffsets[a]) & MAP_SIZE_MASK;
            if (mMap[testOffset] >= branchValue - 4)
                addBranchEraser(testOffset);
        }
        it = mBranchErasers.erase(it);
    }

    it = mNewBranchErasers.begin();
    itEnd = mNewBranchErasers.end();
    for (; it != itEnd; ++it)
        mBranchErasers.insert(*it);
    mNewBranchErasers.clear();

    UINT updatedEraserCount = mBranchErasers.size();

    if (oldEraserCount > 0)
        ++mEraserComboLength;
    if (oldEraserCount > 0 && updatedEraserCount == 0 &&
        mEraserComboLength > BRANCH_BONUS_MULTIPLIER_DIV)
    {
        UINT32 bonus = mEraserComboLength;
        UINT32 multi = bonus / BRANCH_BONUS_MULTIPLIER_DIV;
        bonus *= multi;
        mScore += bonus;
        mEraserComboLengthDisplay = bonus;
        mEraserComboLengthDisplayTicks = BRANCH_SIZE_BONUS_DISPLAY;
        mEraserComboLength = 0;
        mBulletCount += ERASER_COMBO_BULLET_BONUS;
    }

    if (mStuckDLAPs < STUCK_DLAPS_LEVEL_THRESHOLD &&
        updatedEraserCount == 0 &&
        mFrame >= 120)
    {
        mFrame = 0;
        setGameState(STATE_LEVEL_CHANGE);
    }

    if (oldEraserCount > 0 && updatedEraserCount == 0)
        playSoundEffect(SND_ERASER_END);
}


void GameState::updateBullets()
{
    BulletList::iterator it = mBullets.begin();
    BulletList::iterator itEnd = mBullets.end();
    while (it != itEnd)
    {
        Bullet &b = *it;
        b.x += b.xi;
        b.y += b.yi;
        const INT px = (INT)b.x, py = (INT)(MAP_HEIGHT - 1 - b.y);
        const INT32 offset = py * MAP_WIDTH + px;
        if (px >= 0 && py >= 0 && px < MAP_WIDTH && py < MAP_HEIGHT &&
            mMap[offset] > 0)
        {
            assert(offset >= 0 && offset < MAP_SIZE);

            //mBranchErasers.insert(offset);
            UINT a;
            addBranchEraser(offset);
            for (a = 0; a < 8; ++a)
            {
                UINT32 testOffset = (offset + sWanderOffsets[a]) & MAP_SIZE_MASK;
                addBranchEraser(testOffset);
            }

            it = mBullets.erase(it);
        }
        else if (b.x < -MAP_WIDTH / 2 || b.x > 3 * MAP_WIDTH / 2 ||
                 b.y < -MAP_HEIGHT / 2 || b.y > 3 * MAP_HEIGHT / 2)
            it = mBullets.erase(it);
        else
            ++it;
    }
}


void GameState::drawBullets()
{
    BulletList::iterator it = mBullets.begin();
    BulletList::iterator itEnd = mBullets.end();
    for (; it != itEnd; ++it)
    {
        const Bullet &b = *it;
        FLOAT x = b.x * 2 + DLA_X_OFFSET;
        FLOAT y = (b.y - MAP_HEIGHT / 2) * 2 + HEIGHT / 2;
        drawImageRotatedCenteredScaledAdditive(mBulletImage, x, y, b.angle, 1);
    }
}


void GameState::drawHUD()
{
    enter2DMode(WIDTH, HEIGHT);

    drawAimLine();
    drawBullets();
    drawBorders();
    drawBlade();
    drawBorderTexts();

    drawStateTexts();

    leave2DMode();
}


void GameState::drawStateTexts()
{
    if (mFrame < 2)
        return;

    char buf[256];
    FLOAT x = DLA_X_OFFSET + MAP_WIDTH;
    FLOAT y = HEIGHT / 4;
    FLOAT c = sin(mTime * 0.01f) * 0.4f + 0.6f;

    switch (mGameState)
    {
    default:
        break;

    case STATE_LEVEL_CHANGE:
        mBigFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER, "LEVEL DONE", c, 1, c);
        y += mBigFont->getFontHeight();
        mMediumFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, "Bonus Score (Removers x 10)",
                                mBorderR * 2.5f, mBorderG * 2.5f, mBorderB * 2.5f, 0.7f);
        y += mMediumFont->getFontHeight();
        sprintf(buf, "%d", mDisplayLevelBonusScore);
        mBigFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, buf,
                             mBorderR * 2.5f, mBorderG * 2.5f, mBorderB * 2.5f, 0.9f);
        y += mBigFont->getFontHeight();
        mMediumFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, "Final Score", 1, 1, 1, 0.7f);
        y += mMediumFont->getFontHeight();
        sprintf(buf, "%d", mScore);
        mBigFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, buf);
        y += mBigFont->getFontHeight() + mMediumFont->getFontHeight();
        mMediumFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, "Press Space to advance", 1, 1, 1, 0.9f);
        break;

    case STATE_GAME_OVER:
        mBigFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER, "GAME OVER", 1, c, c);
        y += mBigFont->getFontHeight();
        mMediumFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, "Final Score",
                                mBorderR * 2.5f, mBorderG * 2.5f, mBorderB * 2.5f, 0.9f);
        y += mMediumFont->getFontHeight() * 2;
        sprintf(buf, "%d", mScore);
        mBigFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, buf);
        y += mBigFont->getFontHeight() + mMediumFont->getFontHeight() * 2;
        mMediumFont->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER, "Press Space to continue",
                                mBorderR * 2.5f, mBorderG * 2.5f, mBorderB * 2.5f, 0.9f);
        break;
    }
}


void GameState::drawBorderTexts()
{
    char buf[128];
    FLOAT rx = WIDTH - 8, y = 8;

    mMediumFont->drawString(rx, y, UIFont::ALIGN_RIGHT, "Score", 1, 1, 1, 0.75f);
    y += mMediumFont->getFontHeight();
    sprintf(buf, "%d", mScore);
    mNumFont->drawString(rx, y, UIFont::ALIGN_RIGHT, buf);

    y = HEIGHT - 8;
    sprintf(buf, "%d", mBulletCount);
    mNumFont->drawString(rx, y, UIFont::ALIGN_RIGHT | UIFont::ALIGN_BOTTOM, buf);
    y -= mNumFont->getFontHeight();
    mMediumFont->drawString(rx, y, UIFont::ALIGN_RIGHT | UIFont::ALIGN_BOTTOM, "Removers", 1, 1, 1, 0.75f);

    if (mEraserComboLengthDisplayTicks > 0)
    {
        FLOAT alpha = (FLOAT)mEraserComboLengthDisplayTicks / BRANCH_SIZE_BONUS_DISPLAY;
        UIFont *font = mSmallFont;
        y = HEIGHT / 2;
        font->drawString(rx, y, UIFont::ALIGN_RIGHT, "Branch", 1, 1, 1, alpha);
        y += font->getFontHeight();
        font->drawString(rx, y, UIFont::ALIGN_RIGHT, "Size", 1, 1, 1, alpha);
        y += font->getFontHeight();
        font->drawString(rx, y, UIFont::ALIGN_RIGHT, "Bonus", 1, 1, 1, alpha);
        y += font->getFontHeight();
        sprintf(buf, "%d", mEraserComboLengthDisplay);
        mMediumFont->drawString(rx, y, UIFont::ALIGN_RIGHT, buf, 1, 1, 1, alpha);
    }

    FLOAT x = 8;
    y = HEIGHT - 8;
    sprintf(buf, "%d", mLevel);
    mBigFont->drawString(x, y, UIFont::ALIGN_BOTTOM, buf);
    y -= mBigFont->getFontHeight();
    mSmallFont->drawString(x, y, UIFont::ALIGN_BOTTOM, "Level");

#ifdef _DEBUG
    sprintf(buf, "Stuck DLAPs: %d", mStuckDLAPs);
    mSmallFont->drawString(0, 0, 0, buf);
    sprintf(buf, "DLAP radius: %d", mDLAPInitRadius);
    mSmallFont->drawString(0, mSmallFont->getFontHeight(), 0, buf);
#endif
}


void GameState::drawAimLine()
{
    FLOAT distance = MAP_WIDTH;
    FLOAT centerX = DLA_X_OFFSET + MAP_WIDTH, centerY = HEIGHT / 2;
    FLOAT bladeCenterX = centerX + cos(mBladeAngle) * distance;// * 1.04f;
    FLOAT bladeCenterY = centerY + sin(mBladeAngle) * distance;

    glEnable(GL_LINE_STIPPLE);
    glLineStipple(2, 1 << ((mTime / 32) & 15));

    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(1, 1, 1, 0.5f);
    glLineWidth(1);
    glBegin(GL_LINES);
    glVertex2f(bladeCenterX, bladeCenterY);
    glVertex2f(centerX, centerY);
    glEnd();
    glEnable(GL_TEXTURE_2D);
}


static inline void quadFlat(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2)
{
    x2 += 1;
    y2 += 1;
    glBegin(GL_QUADS);
    glVertex2f(x1, y1);
    glVertex2f(x2, y1);
    glVertex2f(x2, y2);
    glVertex2f(x1, y2);
    glEnd();
}


static inline void quadRot(const IMAGE &img, FLOAT x, FLOAT y, UINT rot)
{
    assert(rot < 4);
    glBindTexture(GL_TEXTURE_2D, img.glid);
    FLOAT uv[4];
    INT uvi[4] = { 0, 1, 2, 3 };
    switch (rot)
    {
    case 0: default: break;
    case 2: uvi[0] = 2; uvi[2] = 0; /* fall through */
    case 1: uvi[1] = 3; uvi[3] = 1; break;
    case 3: uvi[0] = 2; uvi[2] = 0; break;
    }
    uv[0] = img.uv[uvi[0]];
    uv[1] = img.uv[uvi[1]];
    uv[2] = img.uv[uvi[2]];
    uv[3] = img.uv[uvi[3]];
    const FLOAT coords[8] = {
        x, y,
        img.w + x, y,
        x, img.h + y,
        img.w + x, img.h + y };
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(uv[0], uv[1]); glVertex2f(coords[0], coords[1]);
    glTexCoord2f(uv[2], uv[1]); glVertex2f(coords[2], coords[3]);
    glTexCoord2f(uv[0], uv[3]); glVertex2f(coords[4], coords[5]);
    glTexCoord2f(uv[2], uv[3]); glVertex2f(coords[6], coords[7]);
    glEnd();
}


void GameState::drawBorders()
{
    glColor4f(mBorderR, mBorderG, mBorderB, 1);
    glDisable(GL_TEXTURE_2D);

    // left
    quadFlat(0, 0, DLA_X_OFFSET, HEIGHT - 1);
    // right
    quadFlat(DLA_X_OFFSET + MAP_WIDTH * 2, 0, WIDTH - 1, HEIGHT - 1);

    glEnable(GL_TEXTURE_2D);
    FLOAT xOffs = DLA_X_OFFSET;
    quadRot(mQuadCircleImage, xOffs, 0, 0);
    quadRot(mQuadCircleImage, xOffs, HEIGHT / 2, 1);
    quadRot(mQuadCircleImage, xOffs + mQuadCircleImage.w, HEIGHT / 2, 2);
    quadRot(mQuadCircleImage, xOffs + mQuadCircleImage.w, 0, 3);
}


void GameState::drawBlade()
{
    FLOAT distance = (MAP_WIDTH + mBladeImage.w / 6) + fabs(sin(mTime * 0.005f)) * 8;
    FLOAT bladeCenterX = DLA_X_OFFSET + MAP_WIDTH, bladeCenterY = HEIGHT / 2;
    bladeCenterX += cos(mBladeAngle) * distance;// * 1.04f;
    bladeCenterY += sin(mBladeAngle) * distance;

    drawImageRotatedCenteredScaledAdditive(mBladeImage, bladeCenterX, bladeCenterY, mBladeAngle, 1);

    if (mFireDelay == 0 && mBulletCount > 0)
    {
        FLOAT distance = MAP_WIDTH - mBulletImage.w / 2;
        FLOAT cx = DLA_X_OFFSET + MAP_WIDTH, cy = HEIGHT / 2;
        cx += cos(mBladeAngle) * distance;// * 1.04f;
        cy += sin(mBladeAngle) * distance;
        drawImageRotatedCenteredScaledAdditive(mBulletImage, cx, cy, mBladeAngle, 1);
    }
}


void GameState::fire()
{
    if (mBulletCount <= 0)
        return;

    Bullet b;
    FLOAT angle = mBladeAngle;
    FLOAT r = MAP_WIDTH * 0.5f;
    FLOAT ca = cos(angle);
    FLOAT sa = sin(angle);
    b.x = (MAP_WIDTH / 2) + r * ca;
    b.y = (MAP_HEIGHT / 2) + r * sa;
    b.xi = -ca * BULLET_SPEED / BULLET_UPDATE_ITERATIONS;
    b.yi = -sa * BULLET_SPEED / BULLET_UPDATE_ITERATIONS;
    b.angle = angle;
    mBullets.push_front(b);

    mFireDelay = mFireDelayAdder;

    --mBulletCount;

    playSoundEffect(SND_SHOOT);
}


void GameState::initDLAParticle(UINT32 n)
{
    assert(n < mDLAPCount);
    FLOAT distance = mDLAPInitRadius;// * mRandom.genrand_real1();
    if (distance >= MAP_WIDTH / 2)
        mDLAPs[n] = mRandom.genrand_int32() & MAP_SIZE_MASK;
    else
    {
        FLOAT angle = 2 * PI * mRandom.genrand_real2();
        INT32 x = (INT32)(cos(angle) * distance + MAP_WIDTH / 2);
        INT32 y = (INT32)(sin(angle) * distance + MAP_HEIGHT / 2);
        if (x < 0)
            x = 0;
        if (x >= MAP_WIDTH)
            x = MAP_WIDTH - 1;
        if (y < 0)
            y = 0;
        if (y >= MAP_HEIGHT)
            y = MAP_HEIGHT - 1;
        mDLAPs[n] = y * MAP_WIDTH + x;
    }
}


BOOL GameState::diffusionLimitedAggregation()
{
    BOOL result = FALSE;
    UINT32 a;
    INT32 maxDist = 0;

    for (a = 0; a < mDLAPCount; ++a)
    {
        const UINT direction = mRandom.genrand_int32() & 7;
        UINT32 offset = mDLAPs[a];
        offset += sWanderOffsets[direction];
        offset &= MAP_SIZE_MASK;
        UINT32 b, hitAccum = 0;//, hitCount = 0;
        for (b = 0; b < 8; ++b)
        {
            const UINT32 testOffset = (offset + sWanderOffsets[b]) & MAP_SIZE_MASK;
            const UINT32 testValue = mMap[testOffset];
            if (testValue)
            {
                //hitAccum += testValue;
                //++hitCount;

                //if (hitAccum < testValue)
                //    hitAccum = testValue;

                hitAccum |= testValue;
            }
        }
        if (hitAccum > 0)
        {
            //mMap[offset] = (hitAccum / hitCount) + 1;
            if (mMap[offset] == 0)
                ++mStuckDLAPs;
            mMap[offset] = hitAccum + 1;
            INT y = (offset >> MAP_WIDTH_BITS) - MAP_HEIGHT / 2;
            INT x = (offset & MAP_WIDTH_MASK) - MAP_WIDTH / 2;
            INT32 dist = (INT32)sqrt(x * x + y * y);
            if (maxDist < dist)
                maxDist = dist;
            if (dist > mDLAPInitRadius)
                mDLAPInitRadius = dist + DLA_MIN_INIT_RADIUS;
            initDLAParticle(a);
            result = TRUE;
        }
        else
            mDLAPs[a] = offset;
    }

    // test small set of DLAPs each frame if they should be reinitialized
    const UINT setDiv = 10;
    UINT setSize = mDLAPCount / setDiv;
    UINT f = mFrame % setDiv;
    for (a = f * setSize; a < (f + 1) * setSize; ++a)
    {
        UINT32 offset = mDLAPs[a];
        UINT y = (offset >> MAP_WIDTH_BITS) - MAP_HEIGHT / 2;
        UINT x = (offset & MAP_WIDTH_MASK) - MAP_WIDTH / 2;
        UINT32 distSq = (INT32)(x * x + y * y);
        const INT32 testRadius = mDLAPInitRadius + DLA_MIN_INIT_RADIUS;
        if (distSq > testRadius * testRadius)
            initDLAParticle(a);
    }

    if (maxDist > MAP_WIDTH / 2)
    {
        mFrame = 0;
        setGameState(STATE_GAME_OVER);
    }

    return result;
}
