/* Copyright 2004 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: GameState.h,v 1.7 2004/10/17 23:11:11 tonic Exp $
 * $Revision: 1.7 $
 */

#ifndef GAMESTATE_H_INCL
#define GAMESTATE_H_INCL


class GameState : public State, public UIEventListener
{
public:

    enum GAME_STATE_ENUM
    {
        STATE_INIT = 0,
        STATE_RUNNING,
        STATE_LEVEL_CHANGE,
        STATE_GAME_OVER
    };

    enum GAME_CONSTANTS_ENUM
    {
        MAP_WIDTH_BITS              = 8,
        MAP_HEIGHT_BITS             = 8,
        MAP_WIDTH                   = (1 << MAP_WIDTH_BITS),
        MAP_HEIGHT                  = (1 << MAP_HEIGHT_BITS),
        MAP_SIZE                    = (MAP_WIDTH * MAP_HEIGHT),
        MAP_WIDTH_MASK              = (MAP_WIDTH - 1),
        MAP_HEIGHT_MASK             = (MAP_HEIGHT - 1),
        MAP_SIZE_MASK               = (MAP_SIZE - 1),
        MAP_IMAGE_WIDTH             = MAP_WIDTH * 2,    // a bit hardcoded multiplier
        MAP_IMAGE_HEIGHT            = 480,
        DLA_MIN_INIT_RADIUS         = 10,
        DLA_X_OFFSET                = 8,
        DLA_PALETTE_STRIDES_GEN_MIN = 8,
        DLA_PALETTE_STRIDES_GEN_MAX = 11,
        DLA_PALETTE_STRIDES         = 7,
        DLA_PALETTE_STRIDE_LENGTH   = 64,
        DLA_PALETTE_LENGTH          = DLA_PALETTE_STRIDES * DLA_PALETTE_STRIDE_LENGTH,
        DLA_ITERATIONS              = 1, // number of DLA iterations per update cycle
        DLAP_COUNT                  = 300,
        DLAP_COUNT_LEVEL_ADD        = 100,
        DLAP_COUNT_MAX              = 1200,
        STUCK_DLAPS_LEVEL_THRESHOLD = 150,
        FIRE_DELAY                  = 10,
        BULLETS_INITIAL_COUNT       = 20,
        BRANCH_SIZE_BONUS_DISPLAY   = 120,
        BRANCH_BONUS_MULTIPLIER_DIV = 30,
        BULLET_REGEN_RATE           = 120,
        BULLET_SPEED                = 1,
        BULLET_UPDATE_ITERATIONS    = 2,
        ERASER_COMBO_BULLET_BONUS   = 5
    };

    GameState(UIGraphics *graphics);
    virtual ~GameState();

    // From State:

    virtual BOOL init();
    virtual void deinit();

    virtual BOOL update(UINT32 time);
    virtual void render(SDL_Surface *screen);

    virtual void onActivate(UINT32 time);
    virtual void onDeactivate();

    virtual void onKeyDown(const SDLKey &key);
    virtual void onKeyUp(const SDLKey &key);

    // GameState:

    void setGameState(INT gameState) { mGameState = gameState; }
    INT getGameState() { return mGameState; }

protected:

    BOOL initUI();
    void deinitUI();

    void startGame();
    void restartGame();
    void deinitGame();

    void updateGameRunningState();
    void makeDLAPalette();
    void initDLAParticle(UINT32 n);
    BOOL diffusionLimitedAggregation();
    void updateBullets();
    void updateBranchErasers();
    void addBranchEraser(UINT32 offset);
    void fire();

    void drawDLA();
    void enhanceDrawnDLA();
    void drawAimLine();
    void drawBullets();
    void drawHUD();
    void drawBorders();
    void drawBlade();
    void drawBorderTexts();
    void drawStateTexts();

    MersenneTwister mRandom;
    UINT32 mTime;
    UINT32 mFrame;
    //UI mUI;
    INT mGameState;
    BOOL mInitialized, mGLInitialized;

    //UIGraphics *mGraphics;
    //SDL_Surface *mMapImage;

    struct Bullet
    {
        FLOAT x, y;
        FLOAT xi, yi;
        FLOAT angle;
    };
    typedef std::list<Bullet> BulletList;
    typedef std::set<UINT32> BranchEraserSet;

    UINT32 *mMap;
    BranchEraserSet mBranchErasers;
    BranchEraserSet mNewBranchErasers;  // added at end of branch eraser update

    typedef UINT32 DLAParticle;
    UINT32 mDLAPInitRadius;
    UINT32 mDLAPCount;
    DLAParticle *mDLAPs; // DLA particle map offsets.
    GLuint mDLATextures[16];
    INT mDLATextureCount;

    BulletList mBullets;

    UINT32 mDLAPalette[DLA_PALETTE_LENGTH];
    FLOAT mBorderR, mBorderG, mBorderB;

    double mBladeAngle, mBladeDestAngle;
    FLOAT mBladeSpeed;
    UINT mFireDelay, mFireDelayAdder;
    UINT mLevel;
    BOOL mKeyLeft, mKeyRight;
    UINT32 mEraserComboLength;  // for branch size bonus
    UINT32 mScore;
    UINT32 mBulletCount;
    UINT32 mEraserComboLengthDisplay, mEraserComboLengthDisplayTicks;
    UINT32 mBulletRegenCount;
    INT32 mStuckDLAPs;
    UINT32 mDisplayLevelBonusScore;

    IMAGE mQuadCircleImage;
    IMAGE mBladeImage;
    IMAGE mBulletImage;

    // UI:

    UIFont *mSmallFont;
    UIFont *mMediumFont;
    UIFont *mBigFont;
    UIFont *mNumFont;
};


#endif // !GAMESTATE_H_INCL
