/* Copyright 2004 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: TitleState.cpp,v 1.3 2004/10/17 21:42:04 tonic Exp $
 * $Revision: 1.3 $
 */

#include "main.h"


TitleState::TitleState() :
    mPlayButton(NULL),
    mEditorButton(NULL),
    mExitButton(NULL),
    mButtonUpImage(NULL),
    mButtonDownImage(NULL)
{
    mFPS = 50;
}


TitleState::~TitleState()
{
    deinit();
}


BOOL TitleState::init()
{
    GameApp *ga = GameApp::getInstance();
    INT scrW = ga->getScreenWidth(), scrH = ga->getScreenHeight();

    UIGraphics *graphics = ga->getGraphics();
    mUI.setGraphics(graphics);

    UIFont *font = ga->getMediumFont();
    UIFont *smallFont = ga->getSmallFont();
    UIFont *bigFont = ga->getBigFont();

    mButtonUpImage = graphics->loadImage(DATA_PREFIX "button_up.png");
    mButtonDownImage = graphics->loadImage(DATA_PREFIX "button_down.png");

    const UINT buttonWidth = mButtonUpImage->getWidth();
    const UINT buttonHeight = mButtonUpImage->getHeight();
    const UINT gap = 50;

    mPlayButton = new UIButton(scrW / 2 - buttonWidth / 2, scrH / 2, mButtonUpImage, mButtonDownImage, NULL, font, "Play");
    mPlayButton->setItemColor(0.8f, 1, 0.8f, 1);
    mPlayButton->setColor(0.8f, 1, 0.8f, 1);
    mPlayButton->addEventListener(this);
    mUI.addUIComponent(mPlayButton);

    mExitButton = new UIButton(scrW / 2 - buttonWidth / 2, scrH / 2 + buttonHeight * 2, mButtonUpImage, mButtonDownImage, NULL, font, "Exit");
    mExitButton->setItemColor(1, 0.8f, 0.8f, 1);
    mExitButton->setColor(1, 0.8f, 0.8f, 1);
    mExitButton->addEventListener(this);
    mUI.addUIComponent(mExitButton);

    const UINT lgap = 8 + bigFont->getFontHeight();
    UILabel *label;
    label = new UILabel(scrW / 2, gap * 2, NULL, bigFont, "Blight Warrior", UIFont::ALIGN_HORIZ_CENTER);
    mUI.addUIComponent(label);
    label = new UILabel(scrW / 2, gap * 2 + lgap, NULL, smallFont,
                        "Version " VERSION_STR, UIFont::ALIGN_HORIZ_CENTER);
    label->setItemColor(0.7f, 0.7f, 0.7f, 1);
    mUI.addUIComponent(label);

    label = new UILabel(scrW / 2, scrH - 1, NULL, smallFont,
                        "Copyright 2004 Jetro Lauha - http://jet.ro",
                        UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_BOTTOM);
    label->setItemColor(0.5f, 0.5f, 0.5f, 1);
    mUI.addUIComponent(label);

    mUI.activateAllComponents();

    return TRUE;
}


void TitleState::deinit()
{
    mUI.deleteAllUIComponents();

    delete mButtonUpImage;
    mButtonUpImage = NULL;
    delete mButtonDownImage;
    mButtonDownImage = NULL;
}


BOOL TitleState::update(UINT32 time)
{
    mTime = time;

    return TRUE;
}


void TitleState::render(SDL_Surface *screen)
{
    clear(0, 0, 0);
    //enter2DMode(WIDTH, HEIGHT);

    GameApp *app = (GameApp *)App::getInstance();

    mUI.render();
    // ...

    //leave2DMode();
}


void TitleState::onKeyDown(const SDLKey &key)
{
    if (key == SDLK_SPACE || key == SDLK_KP_ENTER || key == SDLK_RETURN)
        gameStart();
}


void TitleState::onMouseDown(const SDL_MouseButtonEvent &event)
{
    mUI.handleUIPointerDown(event.x, event.y);
}


void TitleState::onMouseUp(const SDL_MouseButtonEvent &event)
{
    mUI.handleUIPointerUp(event.x, event.y);
}


void TitleState::onMouseMotion(const SDL_MouseMotionEvent &event)
{
    mUI.handleUIPointerMove(event.x, event.y);
}


void TitleState::gameStart()
{
    GameApp *app = (GameApp *)App::getInstance();
    if (app->isAlive())
        app->setState(GameApp::STATE_GAME_RUN);
}


void TitleState::onAction(UIComponent *component)
{
    if (component == mPlayButton)
        gameStart();
    else if (component == mExitButton)
        App::getInstance()->setAlive(FALSE);
}
