/* Copyright 2004 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: sound.cpp,v 1.3 2004/10/17 20:59:44 tonic Exp $
 * $Revision: 1.3 $
 */

#include "main.h"
//#include "CFL.h"
#include "SDL_mixer.h"


//#define DISABLE_SOUND


#define CHANNELS    8
#define MUSICFILE   DATA_PREFIX "ldfungus.it"
#define MUSICVOLMUL 0.8f
#define SFXVOLMUL   1.0f


UINT32 lastSoundEffectTick[SOUNDEFF_TYPES] = {0,};


typedef struct
{
    INT type;
    UINT8 volume;
    UINT8 loop;
    char *filename;
} SOUNDEFF;


static SOUNDEFF sampleData[] =
{
    { SND_ERASER, 0xff, 0, DATA_PREFIX "eraser.ogg" },
    { SND_ERASER_END, 0x90, 0, DATA_PREFIX "eraser_end.ogg" },
    { SND_LEVEL_BEGIN, 0xb0, 0, DATA_PREFIX "level_begin.ogg" },
    { SND_SHOOT, 0x90, 0, DATA_PREFIX "shoot.ogg" },
};
#define SAMPLES (sizeof(sampleData) / sizeof(sampleData[0]))


static BOOL soundInitialized = FALSE, soundEnabled = TRUE, musicEnabled = FALSE;
static Mix_Chunk *samples[SAMPLES];
static Mix_Music *music;
static int musicChn = -1;
static float musicVolume = 1.0f;


static Mix_Chunk * loadSample(char *fileName, CFL *cfl)
{
#ifndef DISABLE_SOUND
    /*if (cfl != NULL)
    {
        Mix_Chunk *res;
        char *data;
        int size;
        data = cfl->getFile(fileName, size);
        SDL_RWops *rw = SDL_RWFromMem((void *)data, size);
        res = Mix_LoadWAV_RW(rw, 1);
        delete[] data;
        return res;
    }
    else*/
    {
        return Mix_LoadWAV(fileName);
    }
#else
    return NULL;
#endif // DISABLE_SOUND
}


static void loadSamples(CFL *cfl)
{
#ifndef DISABLE_SOUND
    int a;

    for (a = 0; a < SAMPLES; ++a)
    {
        samples[a] = loadSample(sampleData[a].filename, cfl);
        if (samples[a] == NULL)
        {
            errorMessage("Sample Load Error", "Error loading sample %s\nError: %s",
                         sampleData[a].filename, SDL_GetError());
            exit(1);
        }
    }
#endif // !DISABLE_SOUND
}


static void loadMusic()
{
#ifndef DISABLE_SOUND
    music = Mix_LoadMUS(MUSICFILE);
    if (music == NULL)
    {
        errorMessage("Music Load Error", "Error loading file %s\nError: %s",
                     MUSICFILE, SDL_GetError());
        exit(1);
    }
#endif // !DISABLE_SOUND
}


void initSound(CFL *cfl)
{
#ifndef DISABLE_SOUND
    if (Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 2048) < 0)
    {
        errorMessage("Sound Disabled", "Unable to initialized audio! (%s)", SDL_GetError());
        soundInitialized = soundEnabled = 0;
        return;
    }
    soundInitialized = 1;
    Mix_AllocateChannels(CHANNELS);

    loadSamples(cfl);
    loadMusic();

    setMusic(1);
#endif // !DISABLE_SOUND
}


void deinitSound()
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    setMusic(0);

    UINT a;
    for (a = 0; a < SAMPLES; ++a)
    {
        Mix_FreeChunk(samples[a]);
        samples[a] = NULL;
    }
    Mix_FreeMusic(music);

    Mix_CloseAudio();
#endif // !DISABLE_SOUND
}


int playSoundEffect(int type, int forceChannel)
{
#ifndef DISABLE_SOUND
    int chn, snd, cnt = 0;

    if (!soundInitialized || !soundEnabled)
        return -1;

    do {
        snd = rand() % SAMPLES;
        ++cnt;
    } while (sampleData[snd].type != type && cnt < 10000);
    if (cnt >= 10000)
        return -1;     // just don't play anything if we can't find a suitable one :)
        //errorMessage("Sound Error", "Unable to pick sample");

    int loop = 0;
    if  (sampleData[snd].loop)
        loop = -1;
    chn = Mix_PlayChannel(forceChannel, samples[snd], loop);

    if (chn >= 0)
        Mix_Volume(chn, (int)(SFXVOLMUL * sampleData[snd].volume / 2));

    return chn;
#else
    return 0;
#endif // DISABLE_SOUND
}


void stopSoundEffect(int channel)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized || channel < 0)
        return;
    Mix_ExpireChannel(channel, 1);
#endif // DISABLE_SOUND
}


void stopAllSoundEffects()
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    int a;
    for (a = 0; a < CHANNELS; ++a)
    {
        if (musicEnabled && musicChn == a)
            continue;   // don't stop music if music is enabled
        stopSoundEffect(a);
    }
#endif // DISABLE_SOUND
}


void setSound(BOOL enabled)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    soundEnabled = enabled ? 1 : 0;
#endif // !DISABLE_SOUND
}


void setMusic(BOOL enabled)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    if (enabled && !musicEnabled)
    {
        // turn music on
        //musicChn = Mix_FadeInMusic(music, -1, 1000);
        musicChn = Mix_PlayMusic(music, -1);
        musicEnabled = musicChn >= 0 ? 1 : 0;   // not sure if this is needed
        setMusicVolume(musicVolume);
    }
    else if (musicEnabled && !enabled)
    {
        // turn music off
        Mix_FadeOutMusic(100);
        musicEnabled = 0;
    }
#endif // !DISABLE_SOUND
}


void setMusicVolume(float volume)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    musicVolume = volume;

    if (musicEnabled)
        Mix_VolumeMusic((int)(volume * MUSICVOLMUL * 128));
#endif // !DISABLE_SOUND
}


BOOL getSoundEnabled()
{
#ifndef DISABLE_SOUND
    return soundEnabled;
#else
    return FALSE;
#endif
}


BOOL getMusicEnabled()
{
#ifndef DISABLE_SOUND
    return musicEnabled;
#else
    return FALSE;
#endif
}
